require 'fog'
require 'guid'
require 'timeout'
require 'puppet/cloudpack'
require 'puppet/cloudpack/progressbar'
require 'puppet/cloudpack/utils'

# This is a required work-around for Ruby 1.9.3p392 (and, probably, other
# versions of Ruby 1.9.3, and pretty certainly 1.9.0 <=> 1.9.2), in that they
# treat any exception class that responds to `to_str` as, well, something that
# should be turned into a string and thrown as `RuntimeError` by
# `Kernel#raise`.
#
# Sadly the RbVmomi fault class is implemented in such a way that it does
# respond to `to_str`, through delegation, and results in the wrong exception
# being thrown at the wrong time.
#
# This is ugly, but I strongly suspect it of being in line with upstream
# authorial intent; see https://github.com/vmware/rbvmomi/issues/17 for the
# issue submitted to track this.
require 'rbvmomi/fault'
class RbVmomi::Fault
  alias_method :upstream_method_missing, :method_missing
  def method_missing(*args)
    super if args.first == 'to_str' or args.first == :to_str
    upsteam_method_missing(*args)
  end
end

module Puppet::CloudPack
  module VMware
    class NotFoundError < RangeError; end
    class InvalidState < RuntimeError; end
    class << self
      def add_timeout_option(action)
        action.option '--timeout=' do
          summary 'Seconds to wait for API calls to complete'
          description <<-EOT
            The maximum number of seconds to wait for API calls to complete.
            Many of the API calls may take considerable time to run.  If
            not specified, this defaults to #{Fog.timeout} seconds.
          EOT
        end
      end

      def add_vmname_option(action)
        action.option '--vmname=', '-n=' do
          required
          summary 'The name of the new VM.'
          description <<-'EOT'
            The name of the new VM to create.  This name must be unique within the
            folder containing the source VM or template.
          EOT
        end
      end

      def add_template_option(action)
        action.option '--template=', '-t=' do
          required
          summary 'The path of the template to clone from.'
          description <<-'EOT'
            The path of the VMware virtual machine template to clone.  A full path
            to the template is expected in the same format as displayed from the
            list action.
          EOT
        end
      end

      def add_wait_option(action)
        action.option '--wait', '-w' do
          summary 'Wait for the clone operation to finish.'
          description <<-'EOT'
            If this option is specified then Puppet will wait for the VMware copy
            operation to finish.  This could be a lengthy process but provides the
            ability to reliably wait until the VM has finished cloning before
            moving on.
          EOT
        end
      end

      def add_wait_for_boot_option(action)
        action.option '--wait-for-boot', '-i' do
          summary 'Wait for the clone to boot and obtain an IP.'
          description <<-'EOT'
            If this option is specified then Puppet will wait for the VMware copy
            to finish and the VM to boot and obtain an IP address on the network.
            This requires the VMware Tools installed in the guest VM.  This option
            implies the --wait option.
          EOT
        end
      end

      def add_create_options(action)
        add_vmname_option(action)
        add_template_option(action)
        add_wait_option(action)
        add_wait_for_boot_option(action)
        add_timeout_option(action)
      end

      def add_bootstrap_options(action)
        add_create_options(action)
        Puppet::CloudPack.add_init_options(action)
      end

      def create_connection(options = {})
        Puppet.notice "Connecting ..."
        connection = Fog::Compute[:vsphere]
        Puppet.notice "Connected to #{connection.vsphere_server} as #{connection.vsphere_username} (API version #{connection.vsphere_rev})"
        connection
      end

      def list(options = {})
        timeout = options[:timeout] ? options[:timeout].to_i : Fog.timeout
        connection = create_connection
        if options[:folder] then
          do_in_progress_bar(
            :timeout => timeout,
            :message => "Locating",
            :notice  => "Finding Virtual Machines in Folder #{options[:folder]}"
          ) do
            servers = connection.servers('folder' => options[:folder])
            servers.collect { |s| s.attributes }
          end
        else
          do_in_progress_bar(
            :timeout => timeout,
            :message => "Locating",
            :notice  => "Finding all Virtual Machines ..."
          ) do
            servers = connection.servers
            # Return an Array of Hash objects
            servers.collect { |s| s.attributes }
          end
        end
      end

      def create(options = {})
        timeout = options[:timeout] ? options[:timeout].to_i : Fog.timeout
        if not options[:template] =~ /^\/Datacenters/
          raise ArgumentError, "The --template option must be a path, for example /Datacenters/Solutions/vm/templates/puppet_template."
        end
        total_steps  = 2
        total_steps += 1 if options[:wait_for_boot]
        current_step = 1
        connection = create_connection

        # We expect only one result from find_vm().
        template = find_vm(
          :path       => options[:template],
          :connection => connection,
          :message    => "Locating (#{current_step}/#{total_steps})"
        ).first
        current_step += 1

        new_vm = do_in_progress_bar(
          :timeout => timeout,
          :message => "Starting (#{current_step}/#{total_steps})",
          :notice  => "Starting the clone process"
        ) do
          template.clone('name' => options[:vmname], 'wait' => options[:wait])
        end
        current_step += 1

        if options[:wait_for_boot] then
          do_in_progress_bar(
            :timeout => timeout,
            :message => "Booting (#{current_step}/#{total_steps})",
            :notice  => "Waiting for the machine to boot and obtain an IP address ..."
          ) do
            new_vm.wait_for(timeout,15) { ipaddress }
          end
        end
        current_step += 1

        { 'status' => 'success' }.merge(filter_server_attributes(new_vm.attributes))
      end

      def bootstrap(options = {})
        server = self.create(options)
        if server['ipaddress'].nil? or server['ipaddress'].empty?
          fail('Unable to perform install action; no ip address returned by create method')
        end
        Puppet::CloudPack.init(server['ipaddress'], options)
      end

      def stop(options = {})
        server = find_and_validate_vm(options[:path])
        do_in_progress_bar(
          :message => "Stopping",
          :notice  => "Stopping Virtual Machine #{server.path}/#{server.name}"
        ) do
          server.stop(:force => options[:force])
        end
      end

      def start(options = {})
        server = find_and_validate_vm(options[:path])
        do_in_progress_bar(
          :message => "Starting",
          :notice => "Starting Virtual Machine #{server.path}/#{server.name}"
        ) do
          server.start
        end
      end

      def terminate(options = {})
        validate_path options[:path]
        timeout = options[:timeout] ? options[:timeout].to_i : Fog.timeout

        retry_exceptions = {
          # we raise this exception if RbVmomi reportrs that the instance is
          # in an invalid state such as in the middle of powering down when
          # the stop command is sent
          InvalidState => 'The VM is in an invalid state, retrying termination.'
        }

        Puppet::CloudPack::Utils.retry_action( :timeout => timeout, :retry_exceptions => retry_exceptions ) do
          server = find_vm(:path => options[:path]).first

          begin
            destroy_options = {}

            if server.is_a_template == false
              if server.power_state =~ /power.*?on/i then
                do_in_progress_bar(
                  :timeout => timeout,
                  :message => "Stopping",
                  :notice  => "The VM is powered on.  Turning it off."
                ) do
                  server.stop(:force => true)
                end
              end
            else
              # the specified path designates a template in which case we need
              # to pass the inventory path so that fog is able to delete it
              destroy_options['inventory_path'] = server.path.split('/').slice(2 .. -1).join('/') << '/' << server.name
            end

            do_in_progress_bar(
              :timeout => timeout,
              :message => "Destroying",
              :notice  => "Destroying Virtual Machine #{server.path}/#{server.name}"
            ) do
              server.destroy(destroy_options)
            end
          rescue RbVmomi::Fault => e
            # since we want to retry only on certain specific RbVmomi exceptions
            # but all the exceptions share the same class - RbVmomi::Fault - we
            # need to check for those on which we want to retry and throw different
            # exceptions instead as we can't include the RbVmomi::Fault in the
            # retry_exception hash passed to the +Puppet::CloudPack::Utils.retry_action+
            # method call because the method would then be retrying on every RbVmomi
            # exception and not just those specific ones we want it to retry on
            fault_class = e.fault.class
            # InvalidPowerState is being returned by vCenter servers, InvalidState by ESXi servers
            if fault_class == RbVmomi::VIM::InvalidPowerState || fault_class == RbVmomi::VIM::InvalidState
              raise InvalidState
            else
              raise e
            end
          end
        end
      end

      def find(options = {})
        validate_path options[:path]
        found = Array.new
        found << find_vm(:path => options[:path]).map { |vm| vm.attributes }
        found.flatten
      end

      private

      # This method validates a path to a VM.
      # It's import to validate the names before you use take any action on a VM.
      # The find_vm method can return multiple VMs if it's passed just a VM name instead
      # of a path. That's why this method exists.
      def validate_path(path)
        # FIXME: This likely needs great improvement.
        # I need to have more examples of real world setups to do a better regex.
        path =~ /^\// or raise "Invalid path. Must be full path of VM"
      end

      # This method looks up and validates a VM. That is it validates that it is
      # a real VM as opposed to a VM template.
      def find_and_validate_vm(path)
        validate_path path
        vm = find_vm(:path => path).first
        vm.is_a_template and raise "Path #{path} designates a VM template, however a real VM is required here."
        vm
      end

      # Filter server attributes which have symbol keys to have string keys
      def filter_server_attributes(attributes)
        attributes.inject({}) { |memo, (k,v)| memo[k.to_s] = v; memo }
      end

      # Find a VM by path.  This is a private method because I suspect there's a much more efficient
      # way to do this and this will be a nice place to refactor.
      # The options hash will initially only support options['path'] but it may make sense to support
      # search filters like EC2 does if we can figure out efficient API calls to VMware.
      def find_vm(options = {})
        # It's really starting to seem like this should be an instance of a class
        # and the connection should be an instance variable.
        connection = options[:connection] || create_connection

        do_in_progress_bar(
          :message => options[:message] || "Locating",
          :notice => "Locating VM at #{options[:path]}"
        ) do
          paths   = options[:path].split('/')
          vm_folder = paths[0...-1].join('/')
          vm_name = paths[-1]

          found = if vm_folder.empty?
            connection.servers.find_all { |s|  s.name == vm_name }
          else
            # This is a semi-optimized strategy to find a VM in a subset of the
            # inventory.
            connection.servers('folder' => vm_folder).find_all { |s|  s.name == vm_name and s.path == vm_folder }
          end

          if found.empty?
            raise NotFoundError, "Could not find VM at #{options[:path]}"
          end

          #Ensure we return an array
          found.to_a
        end
      end

      # Take a block and a timeout and display a progress bar while we're doing our thing
      def do_in_progress_bar(options = {}, &blk)
        timeout = options[:timeout] ? options[:timeout].to_i : Fog.timeout
        start_time = Time.now
        abort_time = start_time + timeout

        Puppet.notice "#{options[:notice]} (Started at #{start_time.strftime("%I:%M:%S %p")})"
        eta_msg = if (timeout <= 120) then
                    "#{timeout} seconds at #{abort_time.strftime("%I:%M:%S %p")}"
                  else
                    "#{timeout / 60} minutes at #{abort_time.strftime("%I:%M %p")}"
                  end
        Puppet.notice "Control will be returned to you in #{eta_msg} if #{options[:message].downcase} is unfinished."

        progress_bar = Puppet::CloudPack::ProgressBar.new(options[:message], timeout)
        progress_mutex = Mutex.new

        progress_thread = Thread.new do
          loop do
            progress = Time.now - start_time
            progress_mutex.synchronize { progress_bar.set progress }
            sleep 0.5
          end
        end

        block_return_value = nil
        begin
          Timeout.timeout(timeout) do
            block_return_value = blk.call
          end
        ensure
          progress_mutex.synchronize { progress_bar.finish; progress_thread.kill }
        end
        end_time = Time.now
        block_return_value
      end
    end
  end
end
